/*
 * 09 Jun 92
 * - Check whether default button is actually *active* when return/enter
 * is typed.  Sheesh.
 * 11 Jun 92
 * - Added flag argument to SkelDlogFilter() allowing caller to determine
 * whether or not standard filter should do return/enter processing.  Since
 * this is now under control of caller, the test is whether the flag is true,
 * rather than whether there's a dialog-specific filter or not.
 * 04 Nov 93
 * - Made the filter function check whether there's an event hook installed
 * and pass the event to it if so, before dealing with the event.
 * 03 Jan 94
 * - Added functions SkelDlogDefaultItem() and SkelDlogCancelItem() for
 * specifying which dialog items are the default and cancel items.  This
 * allows (i) a different item than aDefItem to be used for the OK button
 * and (ii) keyboard equivalents for the Cancel item to be handled
 * automatically by the standard filter.
 * 04 Jan 94
 * - Added function SkelDlogFilterYD() for installing filters that can be used
 * for calls like the System 7 Standard File dialogs CustomGetFile() and
 * CustomPutFile().
 * - Changed type of SkelDlogFilter() from SkelDlogFilterProcPtr to
 * ModalFilterProcPtr.  (They're equivalent, and compiler now treats them
 * that way, so there's no need for SkelDlogFilterProcPtr anymore.)
 * 07 Jan 94
 * - StandardFilter() now only checks keyDown, not autoKey; if the keypress
 * is one it would respond to, the dialog will end up being dismissed and there
 * is no opportunity for autoKey to occur.
 * 12 Jan 94
 * - Stopped using CallPascalB(); the compiler's smart enough now to figure out
 * when Pascal calling conventions need to be used with (*func)(args) syntax.
 * 13 Jan 94
 * - Formerly when a key was typed that maps to a button, the standard filter
 * checked the button to see if it was active.  If so, it flashed the button and
 * returned the item.  Otherwise it returned false.  This leads to the behavior
 * that the key ends up in the current edittext item if there is one when the
 * button is inactive and not otherwise.  Now the filter flashes the button and
 * returns true if the button is active, but turns the event into a null event
 * if the button's inactive so the key doesn't get processed further.
 */

# include	"TransSkel.h"

# define	normalHilite	0
# define	dimHilite		255

# define	returnKey		13
# define	enterKey		3
# define	escapeKey		27

/*
 * Filter function stack. Used so that the filter function mechanism
 * can be reentrant (multiple simultaneous alerts and dialogs).  Maximum
 * reentrancy level is controlled by maxStack.
 *
 * The filter field is a generic pointer because functions of different
 * types are assigned to it.
 *
 * The defaultItem and cancelItem fields are positive to indicate an
 * item to which return/enter and cmd-period/escape should be mapped.
 * They are zero to indicate no key mapping.  defaultItem can be
 * negative to indicate that return/enter should be mapped to the
 * aDefItem item in the dialog/alert record.
 */

# define	maxStack	10


typedef	struct FilterInfo	FilterInfo;

struct FilterInfo
{
	ProcPtr	filter;
	short	defaultItem;
	short	cancelItem;
};


static short	top = -1;
static FilterInfo filterStack[maxStack];


/*
 * Function used by CommonFilter to check whether a dialog item is
 * a button.  If so, flash the button if it's active and return
 * true.  Otherwise return false.
 */

static Boolean
TestDlogButton (DialogPtr d, short item)
{
ControlHandle	ctrl;
Handle	itemHandle;
short	itemType;
Rect	itemRect;

	GetDItem (d, item, &itemType, &itemHandle, &itemRect);
	if (itemType == ctrlItem + btnCtrl)
	{
		ctrl = (ControlHandle) itemHandle;
		if ((**ctrl).contrlHilite == normalHilite)
		{
			SkelFlashButton (ctrl);
			return (true);
		}
	}
	return (false);
}


/*
 * Standard filter function.  The regular alert- or dialog-specific filter
 * is piggybacked onto this.  The standard filter intercepts events that
 * would be ignored otherwise (currently this includes MultiFinder events,
 * as well as update and activate/deactivate events for non-modal-dialog
 * windows) and shoves them through the normal TransSkel event dispatch
 * mechanism.
 *
 * The standard filter also performs key-to-button mapping, returning the
 * default button when return/enter is typed, and/or the cancel button
 * when cmd-period/escape is typed.
 */

static pascal Boolean 
StandardFilter (DialogPtr d, EventRecord *e, short *item)
{
FilterInfo	*info;
short	what = e->what;
char	c;
ControlHandle	ctrl;
Handle	itemHandle;
short	itemType;
Rect	itemRect;
SkelEventHookProcPtr	p;

	if ((p = SkelGetEventHook ()) != (SkelEventHookProcPtr) nil)
	{
		if ((*p) (e))	/* call hook, and return if hook handles event */
			return (true);
	}

	if (what == osEvt)				/* MultiFinder event */
	{
		SkelRouteEvent (e);
		e->what = nullEvent;
		return (false);
	}
	if (what == updateEvt || what == activateEvt)
	{
		/* if the window is registered, SkelRouteEvent() will know what to do */
		if (SkelWindowRegistered ((WindowPtr) e->message))
		{
			SkelRouteEvent (e);		/* handle it and return false; */
			e->what = nullEvent;	/* that's OK because Dialog Mgr */
			return (false);			/* was going to ignore it anyway */
		}
	}

	info = &filterStack[top];

	/*
	 * If event is a key-press, handle key-to-button mapping if necessay.
	 */
	if (what == keyDown)
	{
		c = e->message & charCodeMask;
		if (info->defaultItem != 0)					/* map return/enter */
		{
			if (c == returnKey || c == enterKey)
			{
				if (info->defaultItem < 0)			/* use dialog's default item */
					*item = ((DialogPeek) d)->aDefItem;
				else								/* use specified item */
					*item = info->defaultItem;
				if (TestDlogButton (d, *item))
					return (true);
				e->what = nullEvent;	/* non-active button; ignore event */
			}
		}
		if (info->cancelItem > 0)					/* map cmd-period/escape */
		{
			if (SkelCmdPeriod (e) || c == escapeKey)
			{
				*item = info->cancelItem;
				if (TestDlogButton (d, *item))
					return (true);
				e->what = nullEvent;	/* non-active button; ignore event */
			}
		}
	}
	return (false);
}


/*
 * Run ModalDialog()-style filter.
 *
 * If CommonFilter() doesn't handle the event, pass it to the
 * dialog-specific filter if there is one.
 * NOTE that the latter filter is called as a Pascal function.
 */
 
static pascal Boolean 
ModalFilter (DialogPtr d, EventRecord *e, short *item)
{
FilterInfo	*info;

	if (StandardFilter (d, e, item))
		return (true);

	info = &filterStack[top];

	if (info->filter != nil)
		return ((*(ModalFilterProcPtr) info->filter) (d, e, item));
		/*return (CallPascalB (d, e, item, info->filter));*/
	return (false);
}


/*
 * Run ModalDialog()-style filter that takes a data argument.
 *
 * If CommonFilter() doesn't handle the event, pass it to the
 * dialog-specific filter if there is one.
 * NOTE that the latter filter is called as a Pascal function.
 */
 
static pascal Boolean 
ModalFilterYD (DialogPtr d, EventRecord *e, short *item, void *data)
{
FilterInfo	*info;

	if (StandardFilter (d, e, item))
		return (true);

	info = &filterStack[top];

	if (info->filter != nil)
		return ((*(ModalFilterYDProcPtr) info->filter) (d, e, item, data));
		/*return (CallPascalB (d, e, item, data, info->filter));*/
	return (false);
}


/*
 * Install a ModalDialog()-style filter, or a filter that takes a data
 * argument.
 *
 * In each case, remember the filter specified by the caller, so it can be
 * invoked by the appropriate standard filter later.  Return a pointer to
 * the standard filter, so the caller can pass that to an alert or dialog
 * call that takes a filter function argument.
 *
 * doReturn is true if the standard filter should try to handle return/enter
 * key as a synonym for clicking the default button if that button is active.
 */

pascal ModalFilterProcPtr
SkelDlogFilter (ModalFilterProcPtr filter, Boolean doReturn)
{
FilterInfo	*info;

	info = &filterStack[++top];
	info->filter = filter;
	info->defaultItem = (doReturn ? -1 : 0);
	return (ModalFilter);
}


pascal ModalFilterYDProcPtr
SkelDlogFilterYD (ModalFilterYDProcPtr filter, Boolean doReturn)
{
FilterInfo	*info;

	info = &filterStack[++top];
	info->filter = filter;
	info->defaultItem = (doReturn ? -1 : 0);
	return (ModalFilterYD);
}


/*
 * Remove the last filter function (which reinstalls the previous one
 * as a side effect).  This is called after the alert or dialog has
 * been dismissed.
 */

pascal void
SkelRmveDlogFilter (void)
{
	--top;
}


/*
 * Set default or cancel item for the current filter
 */

pascal void
SkelDlogDefaultItem (short item)
{
	if (top >= 0)
		filterStack[top].defaultItem = item;
}


pascal void
SkelDlogCancelItem (short item)
{
	if (top >= 0)
		filterStack[top].cancelItem = item;
}
